'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var events = require('events');
var path = _interopDefault(require('path'));
var minimost = _interopDefault(require('minimost'));

const removeBrackets = (v) => v.replace(/[<[].+/, '').trim();
const findAllBrackets = (v) => {
    const ANGLED_BRACKET_RE_GLOBAL = /<([^>]+)>/g;
    const SQUARE_BRACKET_RE_GLOBAL = /\[([^\]]+)\]/g;
    const res = [];
    const parse = (match) => {
        let variadic = false;
        let value = match[1];
        if (value.startsWith('...')) {
            value = value.slice(3);
            variadic = true;
        }
        return {
            required: match[0].startsWith('<'),
            value,
            variadic
        };
    };
    let angledMatch;
    while ((angledMatch = ANGLED_BRACKET_RE_GLOBAL.exec(v))) {
        res.push(parse(angledMatch));
    }
    let squareMatch;
    while ((squareMatch = SQUARE_BRACKET_RE_GLOBAL.exec(v))) {
        res.push(parse(squareMatch));
    }
    return res;
};
const getMinimostOptions = (options) => {
    return {
        default: options.reduce((res, option) => {
            if (option.config.default !== undefined) {
                // Only need to set the default value of the first name
                // Since minimist will automatically do the rest for alias names
                res[option.names[0]] = option.config.default;
            }
            return res;
        }, {}),
        boolean: options
            .filter(option => option.isBoolean)
            .reduce((res, option) => {
            return res.concat(option.names);
        }, []),
        alias: options.reduce((res, option) => {
            if (option.names.length > 1) {
                res[option.names[0]] = option.names.slice(1);
            }
            return res;
        }, {})
    };
};
const findLongest = (arr) => {
    return arr.sort((a, b) => {
        return a.length > b.length ? -1 : 1;
    })[0];
};
const padRight = (str, length) => {
    return str.length >= length ? str : `${str}${' '.repeat(length - str.length)}`;
};

class Option {
    constructor(rawName, description, config) {
        this.rawName = rawName;
        this.description = description;
        this.config = Object.assign({}, config);
        let negated = false;
        this.names = removeBrackets(rawName)
            .split(',')
            .map((v) => {
            let name = v.trim().replace(/^-{1,2}/, '');
            if (name.startsWith('no-')) {
                negated = true;
                name = name.replace(/^no-/, '');
            }
            return name;
        });
        if (negated) {
            this.config.default = true;
        }
        if (rawName.includes('<')) {
            this.required = true;
        }
        else if (rawName.includes('[')) {
            this.required = false;
        }
        else {
            // No arg needed, it's boolean flag
            this.isBoolean = true;
        }
    }
}

class Command {
    constructor(rawName, description) {
        this.rawName = rawName;
        this.description = description;
        this.options = [];
        this.aliasNames = [];
        this.name = removeBrackets(rawName);
        this.args = findAllBrackets(rawName);
        this.examples = [];
        this.config = {};
    }
    usage(text) {
        this.usageText = text;
        return this;
    }
    allowUnknownOptions() {
        this.config.allowUnknownOptions = true;
        return this;
    }
    version(version) {
        this.versionNumber = version;
        this.option('-v, --version', 'Display version number');
        return this;
    }
    example(text) {
        this.examples.push(text);
        return this;
    }
    /**
     * Add a option for this command
     * @param rawName Raw option name(s)
     * @param description Option description
     * @param config Option config
     */
    option(rawName, description, config) {
        const option = new Option(rawName, description, config);
        this.options.push(option);
        return this;
    }
    alias(name) {
        this.aliasNames.push(name);
        return this;
    }
    action(callback) {
        this.commandAction = callback;
        return this;
    }
    /**
     * Check if a command name is matched by this command
     * @param name Command name
     */
    isMatched(name) {
        return this.name === name || this.aliasNames.includes(name);
    }
    get isDefaultCommand() {
        return this.name === '' || this.aliasNames.includes('!');
    }
    /**
     * Check if an option is registered in this command
     * @param name Option name
     */
    hasOption(name) {
        return this.options.find(option => {
            return option.names.includes(name);
        });
    }
    outputHelp(config) {
        const version = this.versionNumber || config.versionNumber;
        const sections = [
            {
                body: `${config.bin}${version ? ` v${version}` : ''}`
            }
        ];
        sections.push({
            title: 'Usage',
            body: `  $ ${config.bin} ${this.usageText || this.rawName}`
        });
        if (config.subCommands && config.subCommands.length > 0) {
            const longestCommandName = findLongest(config.subCommands.map(command => command.rawName));
            sections.push({
                title: 'Commands',
                body: config.subCommands
                    .map(command => {
                    return `  ${padRight(command.rawName, longestCommandName.length)}  ${command.description}`;
                })
                    .join('\n')
            });
            sections.push({
                title: `For more info, run any command with the \`--help\` flag`,
                body: config.subCommands
                    .map(command => `  $ ${config.bin} ${command.name} --help`)
                    .join('\n')
            });
        }
        const options = [...this.options, ...(config.globalOptions || [])];
        if (options.length > 0) {
            const longestOptionName = findLongest(options.map(option => option.rawName));
            sections.push({
                title: 'Options',
                body: options
                    .map(option => {
                    return `  ${padRight(option.rawName, longestOptionName.length)}  ${option.description} ${option.config.default === undefined
                        ? ''
                        : `(default: ${option.config.default})`}`;
                })
                    .join('\n')
            });
        }
        if (this.examples.length > 0) {
            sections.push({
                title: 'Examples',
                body: this.examples.map(v => `  ${config.bin} ${v}`).join('\n')
            });
        }
        if (this.helpCallback) {
            this.helpCallback(sections);
        }
        console.log(sections
            .map(section => {
            return section.title
                ? `${section.title}:\n${section.body}`
                : section.body;
        })
            .join('\n\n'));
    }
    outputVersion(bin) {
        if (this.versionNumber) {
            console.log(`${bin}/${this.versionNumber} ${process.platform}-${process.arch} node-${process.version}`);
        }
        return this;
    }
    checkUnknownOptions(options, globalCommand) {
        if (!this.config.allowUnknownOptions) {
            for (const name of Object.keys(options)) {
                if (name !== '--' &&
                    !this.hasOption(name) &&
                    !globalCommand.hasOption(name)) {
                    console.error(`error: Unknown option \`${name.length > 1 ? `--${name}` : `-${name}`}\``);
                    process.exitCode = 1;
                    return true;
                }
            }
        }
        return false;
    }
}

const NAME_OF_GLOBAL_COMMAND = 'this-does-not-matter';
class CAC extends events.EventEmitter {
    constructor() {
        super();
        this.commands = [];
        this.globalCommand = new Command(NAME_OF_GLOBAL_COMMAND, 'The global command');
        this.globalCommand.usage('<command> [options]');
    }
    usage(text) {
        this.globalCommand.usage(text);
        return this;
    }
    command(rawName, description) {
        const command = new Command(rawName, description);
        this.commands.push(command);
        return command;
    }
    option(rawName, description, config) {
        this.globalCommand.option(rawName, description, config);
        return this;
    }
    help(callback) {
        this.globalCommand.option('-h, --help', 'Display this message');
        this.globalCommand.helpCallback = callback;
        return this;
    }
    version(version) {
        this.globalCommand.version(version);
        return this;
    }
    /**
     * Add a global example
     * @param example
     */
    example(example) {
        this.globalCommand.example(example);
        return this;
    }
    outputHelp() {
        if (this.matchedCommand) {
            this.matchedCommand.outputHelp({
                bin: this.bin,
                subCommands: this.matchedCommand.name === '' ? this.commands : undefined,
                versionNumber: this.globalCommand.versionNumber,
                globalOptions: this.globalCommand.options
            });
        }
        else {
            this.globalCommand.outputHelp({
                bin: this.bin,
                subCommands: this.commands
            });
        }
        return this;
    }
    outputVersion() {
        this.globalCommand.outputVersion(this.bin);
        return this;
    }
    parse(argv = process.argv) {
        this.rawArgs = argv;
        this.bin = argv[1] ? path.basename(argv[1]) : 'cli';
        for (const command of this.commands) {
            const minimostOptions = getMinimostOptions([
                ...this.globalCommand.options,
                ...command.options
            ]);
            const parsed = this.minimost(argv.slice(2), minimostOptions);
            if (command.isMatched(parsed.args[0])) {
                this.matchedCommand = command;
                this.args = parsed.args;
                this.options = parsed.options;
                this.emit(`command:${parsed.args[0]}`, command);
                this.runCommandAction(command, this.globalCommand, parsed);
                return parsed;
            }
        }
        // Try the default command
        for (const command of this.commands) {
            if (command.name === '') {
                const minimostOptions = getMinimostOptions([
                    ...this.globalCommand.options,
                    ...command.options
                ]);
                const parsed = this.minimost(argv.slice(2), minimostOptions);
                this.args = parsed.args;
                this.options = parsed.options;
                this.matchedCommand = command;
                this.emit(`command:!`, command);
                this.runCommandAction(command, this.globalCommand, parsed);
                return parsed;
            }
        }
        const globalMinimostOptions = getMinimostOptions(this.globalCommand.options);
        const parsed = this.minimost(argv.slice(2), globalMinimostOptions);
        this.args = parsed.args;
        this.options = parsed.options;
        if (parsed.options.help && this.globalCommand.hasOption('help')) {
            this.outputHelp();
            return parsed;
        }
        if (parsed.options.version &&
            this.globalCommand.hasOption('version') &&
            this.globalCommand.versionNumber) {
            this.outputVersion();
            return parsed;
        }
        this.emit('command:*');
        return parsed;
    }
    minimost(argv, minimostOptions) {
        const { input: args, flags: options } = minimost(argv, minimostOptions);
        return {
            args,
            options
        };
    }
    runCommandAction(command, globalCommand, { args, options }) {
        if (options.help && globalCommand.hasOption('help')) {
            return this.outputHelp();
        }
        if (options.version && globalCommand.hasOption('version')) {
            return this.outputVersion();
        }
        if (!command.commandAction)
            return;
        if (command.checkUnknownOptions(options, globalCommand))
            return;
        // The first one is command name
        if (!command.isDefaultCommand) {
            args = args.slice(1);
        }
        const minimalArgsCount = command.args.filter(arg => arg.required).length;
        if (args.length < minimalArgsCount) {
            console.error(`error: missing required args for command "${command.rawName}"`);
            process.exitCode = 1;
            return;
        }
        const actionArgs = [];
        command.args.forEach((arg, index) => {
            if (arg.variadic) {
                actionArgs.push(args.slice(index));
            }
            else {
                actionArgs.push(args[index]);
            }
        });
        actionArgs.push(options);
        return command.commandAction.apply(this, actionArgs);
    }
}
const cac = () => new CAC();
module.exports = cac;
module.exports.default = cac;

module.exports = cac;
